import functools
import http.server
import logging
import os
import pathlib
import subprocess
import sys
import tempfile
import threading
from typing import Any
from typing import cast
from typing import TYPE_CHECKING
import unittest

import libtorrent as lt

from . import tdummy

if TYPE_CHECKING:
    from libtorrent import _Entry


class Handler(http.server.BaseHTTPRequestHandler):
    def __init__(self, *args: Any, serve_data: bytes, **kwargs: Any) -> None:
        self.serve_data = serve_data
        super().__init__(*args, **kwargs)

    def do_GET(self) -> None:
        logging.info("%s", self.requestline)
        self.send_response(200)
        self.send_header("content-length", str(len(self.serve_data)))
        self.end_headers()
        self.wfile.write(self.serve_data)


class TestSimpleClient(unittest.TestCase):
    def setUp(self) -> None:
        # path relative to this file
        self.script_path = pathlib.Path(__file__).parent / ".." / "simple_client.py"

        self.tempdir = tempfile.TemporaryDirectory()
        self.tempdir_path = pathlib.Path(self.tempdir.name)

        self.torrent = tdummy.get_default()

        # set up a web seed to serve dummy torrent data
        self.server = http.server.HTTPServer(
            ("127.0.0.1", 0), functools.partial(Handler, serve_data=self.torrent.data)
        )
        addr = self.server.server_address[0]
        port = self.server.server_address[1]
        self.server_thread = threading.Thread(target=self.server.serve_forever)
        self.server_thread.start()

        # construct a .torrent to point to our web seed
        tdict = self.torrent.dict
        tdict[b"url-list"] = b"".join(
            [b"http://", str(addr).encode(), b":", str(port).encode()]
        )
        self.tdict = cast("_Entry", tdict)
        self.torrent_path = self.tempdir_path / "dummy.torrent"
        self.torrent_path.write_bytes(lt.bencode(self.tdict))

    def tearDown(self) -> None:
        self.server.shutdown()
        self.server.server_close()
        self.tempdir.cleanup()

    def test_download_from_web_seed(self) -> None:
        proc = subprocess.run(
            [sys.executable, str(self.script_path), str(self.torrent_path)],
            cwd=self.tempdir_path,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            check=True,
            universal_newlines=True,
            timeout=20,
        )

        self.assertEqual(proc.stderr, "")
        file_path = self.tempdir_path / os.fsdecode(self.torrent.files[0].path)
        self.assertEqual(file_path.read_bytes(), self.torrent.data)



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
from __future__ import print_function

import sys
import os
import resource
import shutil
import shlex
import time
import subprocess
import random
import signal
import hashlib

# this is a disk I/O benchmark script. It runs benchmarks
# over different number of peers.

# to set up the test, build the example directory in release
# and stage client_test and connection_tester to the examples directory:
#
#   bjam link=static release debug-symbols=on stage
#
# make sure gnuplot is installed.

# the following lists define the space tests will be run in

peers = [50, 200, 500, 1000]
# builds = ['rtorrent', 'utorrent', 'libtorrent']
builds = ['libtorrent']

# the number of peers for the filesystem test. The
# idea is to stress test the filesystem by using a lot
# of peers, since each peer essentially is a separate
# read location on the platter
default_peers = peers[1]

# the amount of cache for the filesystem test
# 5.5 GiB of cache
default_cache = 400000

# the number of seconds to run each test. It's important that
# this is shorter than what it takes to finish downloading
# the test torrent, since then the average rate will not
# be representative of the peak anymore
# this has to be long enough to download a full copy
# of the test torrent. It's also important for the
# test to be long enough that the warming up of the
# disk cache is not a significant part of the test,
# since download rates will be extremely high while downloading
# into RAM
test_duration = 100

utorrent_version = 'utorrent-server-alpha-v3_3'

# make sure the environment is properly set up
try:
    if os.name == 'posix':
        resource.setrlimit(resource.RLIMIT_NOFILE, (4000, 5000))
except Exception:
    if resource.getrlimit(resource.RLIMIT_NOFILE)[0] < 4000:
        print('please set ulimit -n to at least 4000')
        sys.exit(1)


def build_stage_dirs():
    ret = []
    for i in builds[2:3]:
        ret.append('stage_%s' % i)
    return ret


# make sure we have all the binaries available
binaries = ['client_test', 'connection_tester']
for b in build_stage_dirs():
    for i in binaries:
        p = os.path.join(b, i)
        if not os.path.exists(p):
            print('make sure "%s" is available in ./%s' % (i, b))
            sys.exit(1)

# make sure we have a test torrent
if not os.path.exists('test.torrent'):
    print('generating test torrent')
    # generate a 100 GB torrent, to make sure it won't all fit in physical RAM
    os.system('./connection_tester gen-torrent -s 100000 -t test.torrent')

# use a new port for each test to make sure they keep working
# this port is incremented for each test run
port = 10000 + random.randint(0, 40000)

try:
    os.mkdir('benchmark-dir')
except Exception:
    pass


def clear_caches():
    if 'linux' in sys.platform:
        os.system('sync')
        try:
            open('/proc/sys/vm/drop_caches', 'w').write('3')
        except Exception:
            pass
    elif 'darwin' in sys.platform:
        os.system('purge')


def build_utorrent_commandline(config, port):
    num_peers = config['num-peers']
    torrent_path = config['torrent']
    target_folder = build_target_folder(config)

    try:
        os.mkdir('utorrent_session')
    except Exception:
        pass
    with open('utorrent_session/settings.dat', 'w+') as cfg:

        cfg.write('d')
        cfg.write('20:ul_slots_per_torrenti%de' % num_peers)
        cfg.write('17:conns_per_torrenti%de' % num_peers)
        cfg.write('14:conns_globallyi%de' % num_peers)
        cfg.write('9:bind_porti%de' % port)
        cfg.write('19:dir_active_download%d:%s' % (len(config['save-path']),
                                                   config['save-path']))
        cfg.write('19:diskio.sparse_filesi1e')
        cfg.write('14:cache.overridei1e')
        cfg.write('19:cache.override_sizei%de' % int(config['cache-size'] *
                                                     16 / 1024))
        cfg.write('17:dir_autoload_flagi1e')
        cfg.write('12:dir_autoload8:autoload')
        cfg.write('11:logger_maski4294967295e')
        cfg.write('1:vi0e')
        cfg.write('12:webui.enablei1e')
        cfg.write('19:webui.enable_listeni1e')
        cfg.write('14:webui.hashword20:' + hashlib.sha1(
            'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaadmin').digest())
        cfg.write('10:webui.porti8080e')
        cfg.write('10:webui.salt32:aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa')
        cfg.write('14:webui.username5:admin')
        cfg.write('e')

    try:
        os.mkdir('utorrent_session/autoload')
    except Exception:
        pass
    try:
        shutil.copy(torrent_path, 'utorrent_session/autoload/')
    except Exception:
        pass
    return './%s/utserver -logfile %s/client.log -settingspath ' % \
        (utorrent_version, target_folder) + \
        'utorrent_session'


def build_rtorrent_commandline(config, port):
    num_peers = config['num-peers']
    torrent_path = config['torrent']
    target_folder = build_target_folder(config)

    if os.path.exists(target_folder):
        add_command = ''
    else:
        try:
            os.mkdir(target_folder)
        except Exception:
            pass
        # it seems rtorrent may delete the original torrent when it's being added
        try:
            shutil.copy(torrent_path, target_folder)
        except Exception:
            pass
        add_command = '-O load_start_verbose=%s/%s ' % (target_folder, torrent_path)

    return ('rtorrent -d %s -n -p %d-%d -O max_peers=%d -O max_uploads=%d %s -s '
            '%s -O max_memory_usage=128000000000') % (
                config['save-path'], port, port, num_peers, num_peers, add_command, target_folder)


def build_libtorrent_commandline(config, port):
    num_peers = config['num-peers']
    torrent_path = config['torrent']
    target_folder = build_target_folder(config)

    return ('./client_test -k -O -F 500 --enable_upnp=0 --enable_natpmp=0 '
            '--enable_dht=0 --mixed_mode_algorithm=0 --peer_timeout=%d '
            '--listen_queue_size=%d --unchoke_slots_limit=%d -T %d '
            '--connections_limit=%d --cache_size=%d -s "%s" '
            '--listen_interfaces="0.0.0.0:%d" --aio_threads=%d '
            '-f %s/client.log %s') % (
                test_duration, num_peers, num_peers, num_peers, num_peers, config['cache-size'],
                config['save-path'], port, config['disk-threads'], target_folder, torrent_path)


def build_commandline(config, port):

    if config['build'] == 'utorrent':
        return build_utorrent_commandline(config, port)

    if config['build'] == 'rtorrent':
        return build_rtorrent_commandline(config, port)

    if config['build'] == 'libtorrent':
        return build_libtorrent_commandline(config, port)


def delete_files(files):
    for i in files:
        print('deleting %s' % i)
        try:
            os.remove(i)
        except Exception:
            try:
                shutil.rmtree(i)
            except Exception:
                try:
                    if os.path.exists(i):
                        print('failed to delete %s' % i)
                except Exception:
                    pass


def build_test_config(num_peers=default_peers, cache_size=default_cache,
                      test='download', build='libtorrent', profile='', disk_threads=16,
                      torrent='test.torrent', disable_disk=False):
    config = {'test': test, 'save-path': os.path.join('.', 'benchmark-dir'), 'num-peers': num_peers,
              'cache-size': cache_size, 'build': build, 'profile': profile,
              'disk-threads': disk_threads, 'torrent': torrent, 'disable-disk': disable_disk}
    return config


def build_target_folder(config):

    no_disk = ''
    if config['disable-disk']:
        no_disk = '_no-disk'

    return 'results_%s_%s_%d_%d_%d%s' % (config['build'],
                                         config['test'],
                                         config['num-peers'],
                                         config['cache-size'],
                                         config['disk-threads'],
                                         no_disk)


def find_library(name):
    paths = ['/usr/lib64/', '/usr/local/lib64/', '/usr/lib/', '/usr/local/lib/']

    for p in paths:
        try:
            if os.path.exists(p + name):
                return p + name
        except Exception:
            pass
    return name


def find_binary(names):
    paths = ['/usr/bin/', '/usr/local/bin/']
    for n in names:
        for p in paths:
            try:
                if os.path.exists(p + n):
                    return p + n
            except Exception:
                pass
    return names[0]


def run_test(config):

    j = os.path.join

    target_folder = build_target_folder(config)
    if os.path.exists(target_folder):
        print('results already exists, skipping test (%s)' % target_folder)
        return

    print('\n\n*********************************')
    print('*          RUNNING TEST         *')
    print('*********************************\n\n')
    print('%s %s' % (config['build'], config['test']))

    # make sure any previous test file is removed
    # don't clean up unless we're running a download-test, so that we leave the test file
    # complete for a seed test.
    delete_files(['utorrent_session/settings.dat', 'utorrent_session/settings.dat.old', 'asserts.log'])
    if config['test'] == 'download' or config['test'] == 'dual':
        delete_files([j(config['save-path'], 'test'),
                      '.ses_state',
                      j(config['save-path'], '.resume'),
                      'utorrent_session',
                      '.dht_state',
                      'rtorrent_session'])

    try:
        os.mkdir(target_folder)
    except Exception:
        pass

    # save off the command line for reference
    global port
    cmdline = build_commandline(config, port)
    binary = cmdline.split(' ')[0]
    environment = None
    if config['profile'] == 'tcmalloc':
        environment = {'LD_PRELOAD': find_library('libprofiler.so.0'),
                       'CPUPROFILE': j(target_folder, 'cpu_profile.prof')}
    if config['profile'] == 'memory':
        environment = {'LD_PRELOAD': find_library('libprofiler.so.0'),
                       'HEAPPROFILE': j(target_folder, 'heap_profile.prof')}
    if config['profile'] == 'perf':
        cmdline = 'perf record -g --output=' + \
            j(target_folder, 'perf_profile.prof') + ' ' + cmdline
    with open(j(target_folder, 'cmdline.txt'), 'w+') as f:
        f.write(cmdline)

    with open(j(target_folder, 'config.txt'), 'w+') as f:
        print(config, file=f)

    print('clearing disk cache')
    clear_caches()
    print('OK')
    client_output = open(j(target_folder, 'client.output'), 'w+')
    client_error = open(j(target_folder, 'client.error'), 'w+')
    print('launching: %s' % cmdline)
    client = subprocess.Popen(
        shlex.split(cmdline),
        stdout=client_output,
        stdin=subprocess.PIPE,
        stderr=client_error,
        env=environment)
    print('OK')
    # enable disk stats printing
    if config['build'] == 'libtorrent':
        print('x', end=' ', file=client.stdin)
    time.sleep(4)
    test_dir = 'upload' if config['test'] == 'download' else 'download' if config['test'] == 'upload' else 'dual'
    cmdline = './connection_tester %s -c %d -d 127.0.0.1 -p %d -t %s' % (
        test_dir, config['num-peers'], port, config['torrent'])
    print('launching: %s' % cmdline)
    tester_output = open(j(target_folder, 'tester.output'), 'w+')
    tester = subprocess.Popen(shlex.split(cmdline), stdout=tester_output)
    print('OK')

    time.sleep(2)

    print('\n')
    i = 0
    while True:
        time.sleep(1)
        tester.poll()
        if tester.returncode is not None:
            print('tester terminated')
            break
        client.poll()
        if client.returncode is not None:
            print('client terminated')
            break
        print('\r%d / %d\x1b[K' % (i, test_duration), end=' ')
        sys.stdout.flush()
        i += 1
        # in download- and dual tests, connection_tester will exit once the
        # client is done downloading. In upload tests, we'll upload for
        # 'test_duration' number of seconds until we end the test
        if config['test'] != 'download' and config['test'] != 'dual' and i >= test_duration:
            break
    print('\n')

    if client.returncode is None:
        try:
            print('killing client')
            client.send_signal(signal.SIGINT)
        except Exception:
            pass

    time.sleep(10)
    client.wait()
    tester.wait()
    tester_output.close()
    client_output.close()
    terminate = False
    if tester.returncode != 0:
        print('tester returned %d' % tester.returncode)
        terminate = True
    if client.returncode != 0:
        print('client returned %d' % client.returncode)
        terminate = True

    try:
        shutil.copy('asserts.log', target_folder)
    except Exception:
        pass

    os.chdir(target_folder)

    if config['build'] == 'libtorrent':
        # parse session stats
        print('parsing session log')
        os.system('python ../../tools/parse_session_stats.py client.log')

    os.chdir('..')

    if config['profile'] == 'tcmalloc':
        print('analyzing CPU profile [%s]' % binary)
        os.system('%s --pdf %s %s/cpu_profile.prof >%s/cpu_profile.pdf' %
                  (find_binary(['google-pprof', 'pprof']), binary, target_folder, target_folder))
    if config['profile'] == 'memory':
        for i in range(1, 300):
            profile = j(target_folder, 'heap_profile.prof.%04d.heap' % i)
            try:
                os.stat(profile)
            except Exception:
                break
            print('analyzing heap profile [%s] %d' % (binary, i))
            os.system('%s --pdf %s %s >%s/heap_profile_%d.pdf' %
                      (find_binary(['google-pprof', 'pprof']), binary, profile, target_folder, i))
    if config['profile'] == 'perf':
        print('analyzing CPU profile [%s]' % binary)
        os.system(('perf report --input=%s/perf_profile.prof --threads --demangle --show-nr-samples '
                   '>%s/profile.txt' % (target_folder, target_folder)))

    port += 1

    if terminate:
        sys.exit(1)


for b in builds:
    for test in ['upload', 'download', 'dual']:
        config = build_test_config(build=b, test=test, profile='perf')
        run_test(config)

for p in peers:
    for test in ['upload', 'download', 'dual']:
        config = build_test_config(num_peers=p, test=test, profile='perf')
        run_test(config)



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4

"""
Test libtorrent disk I/O on various Linux filesystems.

Creates a 350 MiB disk image per filesystem, formats it once, then mounts it
and runs test_filesystem with cwd set to the mount point. Images are preserved
across runs to avoid repeated mkfs overhead.

Requires sudo access for losetup, mount, umount, chmod, and zpool.
"""

from argparse import ArgumentParser
import os
from pathlib import Path
import shutil
import subprocess
import sys
from typing import Optional
from typing import Protocol

SCRIPT_DIR = Path(__file__).parent.resolve()
REPO_ROOT = SCRIPT_DIR.parent
TEST_DIR = REPO_ROOT / "test"
DEFAULT_IMAGE_DIR = SCRIPT_DIR / "fs_images"
MOUNT_BASE = Path("/tmp/lt_fs_test")
IMAGE_SIZE_MB = 350


def header(msg: str) -> None:
    print(f"\n{'='*5} {msg}\n")


def run(
    cmd: list[str], cwd: Optional[Path] = None
) -> subprocess.CompletedProcess[bytes]:
    print(f"  $ {' '.join(cmd)}")
    return subprocess.run(cmd, check=True, cwd=cwd)


def run_sudo(cmd: list[str]) -> subprocess.CompletedProcess[bytes]:
    return run(["sudo"] + cmd)


def attach_loop(image: Path) -> str:
    cmd = ["sudo", "losetup", "--find", "--show", str(image)]
    print(f"  $ {' '.join(cmd)}")
    result = subprocess.run(
        cmd,
        check=True,
        capture_output=True,
        text=True,
    )
    loop_dev = result.stdout.strip()
    # Wait for udev to finish processing the new loop device before we mount
    # it. Without this, udisks2 may still be probing the device when mount(8)
    # tries to open it, causing mount to block indefinitely in kernel.
    run_sudo(["udevadm", "settle"])
    return loop_dev


def detach_loop(loop_dev: str) -> None:
    run_sudo(["losetup", "-d", loop_dev])


class Filesystem(Protocol):
    name: str

    def available(self) -> bool: ...
    def format(self, image: Path) -> None: ...
    def mount(self, image: Path, mount_point: Path) -> None: ...
    def umount(self, mount_point: Path) -> None: ...


class LoopFilesystem:
    """Filesystem that uses a loop device with mkfs + mount."""

    def __init__(
        self, name: str, mkfs_cmd: list[str], mount_opts: Optional[list[str]] = None
    ) -> None:
        self.name = name
        self.mkfs_cmd = mkfs_cmd
        self.mount_opts = mount_opts or []
        self._loop_dev: Optional[str] = None

    def available(self) -> bool:
        return shutil.which(self.mkfs_cmd[0]) is not None

    def format(self, image: Path) -> None:
        loop_dev = attach_loop(image)
        try:
            run_sudo(self.mkfs_cmd + [loop_dev])
        finally:
            detach_loop(loop_dev)

    def mount(self, image: Path, mount_point: Path) -> None:
        loop_dev = attach_loop(image)
        self._loop_dev = loop_dev
        opts = ["-o", ",".join(self.mount_opts)] if self.mount_opts else []
        try:
            run_sudo(["mount"] + opts + [loop_dev, str(mount_point)])
        except Exception:
            detach_loop(loop_dev)
            self._loop_dev = None
            raise

    def umount(self, mount_point: Path) -> None:
        try:
            run_sudo(["umount", str(mount_point)])
        finally:
            if self._loop_dev is not None:
                detach_loop(self._loop_dev)
                self._loop_dev = None


class ZfsFilesystem:
    """ZFS: zpool create both formats and mounts; zpool destroy unmounts."""

    POOL_NAME = "lt_test_zfs"
    name = "zfs"

    def available(self) -> bool:
        return shutil.which("zpool") is not None

    def format(self, image: Path) -> None:
        pass  # zpool create handles formatting and mounting together

    def mount(self, image: Path, mount_point: Path) -> None:
        # Destroy any stale pool left by a previous crashed run
        cmd = ["sudo", "zpool", "destroy", self.POOL_NAME]
        print(f"  $ {' '.join(cmd)}")
        subprocess.run(cmd, capture_output=True)
        run_sudo(
            [
                "zpool",
                "create",
                "-f",
                "-m",
                str(mount_point),
                self.POOL_NAME,
                str(image),
            ]
        )

    def umount(self, mount_point: Path) -> None:
        run_sudo(["zpool", "destroy", self.POOL_NAME])


FILESYSTEMS: list[Filesystem] = [
    LoopFilesystem("ext4", ["mkfs.ext4", "-F"]),  # apt install e2fsprogs
    LoopFilesystem("btrfs", ["mkfs.btrfs", "-f"]),  # apt install btrfs-progs
    # TODO: xfs keeps hanging in the kernel
    # LoopFilesystem("xfs", ["mkfs.xfs", "-f"]),  # apt install xfsprogs
    LoopFilesystem("f2fs", ["mkfs.f2fs", "-f"]),  # apt install f2fs-tools
    LoopFilesystem("jfs", ["mkfs.jfs", "-q"]),  # apt install jfsutils
    LoopFilesystem("nilfs2", ["mkfs.nilfs2", "-f"]),  # apt install nilfs-tools
    LoopFilesystem("bcachefs", ["mkfs.bcachefs", "-f"]),  # apt install bcachefs-tools
    ZfsFilesystem(),  # apt install zfsutils-linux
]


def test_filesystem(
    fs: Filesystem, binary: Path, image_dir: Path, reformat: bool
) -> Optional[bool]:
    """Returns True=pass, False=fail, None=skipped."""
    if not fs.available():
        print(f"[{fs.name}] SKIP  (required tool not found)")
        return None

    header(f"Filesystem: {fs.name}")
    image = image_dir / f"{fs.name}.img"
    mount_point = MOUNT_BASE / fs.name

    if reformat and image.exists():
        print(f"  Removing existing image for reformat: {image}")
        image.unlink()

    if not image.exists():
        print(f"  Creating {IMAGE_SIZE_MB} MiB image: {image}")
        image.parent.mkdir(parents=True, exist_ok=True)
        if shutil.which("fallocate"):
            run(["fallocate", "-l", f"{IMAGE_SIZE_MB}M", str(image)])
        else:
            run(
                [
                    "dd",
                    "if=/dev/zero",
                    f"of={image}",
                    "bs=1M",
                    f"count={IMAGE_SIZE_MB}",
                    "status=none",
                ]
            )
        print(f"  Formatting as {fs.name}...")
        fs.format(image)
    else:
        print(f"  Reusing existing image: {image}")

    mount_point.mkdir(parents=True, exist_ok=True)
    fs.mount(image, mount_point)
    try:
        run_sudo(["chmod", "777", str(mount_point)])
        result = subprocess.run([str(binary)], cwd=mount_point)
        return result.returncode == 0
    finally:
        fs.umount(mount_point)


def main() -> None:
    p = ArgumentParser(
        description="Test libtorrent disk I/O on various Linux filesystems."
    )
    p.add_argument(
        "--reformat", action="store_true", help="Delete and recreate all disk images"
    )
    p.add_argument(
        "--filesystems",
        nargs="+",
        metavar="FS",
        help="Which filesystems to test (default: all supported)",
    )
    args = p.parse_args()

    print("NOTE: sudo is required for losetup / mount / umount / zpool.")

    if os.geteuid() == 0:
        print("WARNING: running as root; images will be owned by root.")

    all_fs_names = [fs.name for fs in FILESYSTEMS]
    if args.filesystems:
        unknown = set(args.filesystems) - set(all_fs_names)
        if unknown:
            sys.exit(
                f"ERROR: unknown filesystem(s): {', '.join(sorted(unknown))}\n"
                f"  Available: {', '.join(all_fs_names)}"
            )
        requested = args.filesystems
    else:
        requested = all_fs_names

    header("Building test_filesystem (link=static)")
    run(["b2", "link=static", "stage_test_filesystem"], cwd=TEST_DIR)
    binary = TEST_DIR / "test_filesystem"
    if not binary.exists():
        sys.exit(f"ERROR: build succeeded but {binary} not found")
    print(f"  Binary: {binary}")

    results: dict[str, Optional[bool]] = {}
    for name in requested:
        fs = next(f for f in FILESYSTEMS if f.name == name)
        results[name] = test_filesystem(fs, binary, DEFAULT_IMAGE_DIR, args.reformat)

    header("Results")
    width = max(len(n) for n in results) + 2
    any_failed = False
    for name, outcome in results.items():
        if outcome is None:
            label = "SKIP"
        elif outcome:
            label = "PASS"
        else:
            label = "FAIL"
            any_failed = True
        print(f"  {name:<{width}} {label}")

    sys.exit(1 if any_failed else 0)


if __name__ == "__main__":
    main()



# mypy: disable-error-code="unreachable"
import functools
import http.server
import logging
import os
import pathlib
import subprocess
import sys
import tempfile
import threading
from typing import Any
from typing import cast
from typing import TYPE_CHECKING
import unittest

import libtorrent as lt

from . import lib
from . import tdummy

if TYPE_CHECKING:
    from libtorrent import _Entry

# import fails on windows
if os.name != "nt":
    import pty


class Handler(http.server.BaseHTTPRequestHandler):
    def __init__(self, *args: Any, serve_data: bytes, **kwargs: Any) -> None:
        self.serve_data = serve_data
        super().__init__(*args, **kwargs)

    def do_GET(self) -> None:
        logging.info("%s", self.requestline)
        self.send_response(200)
        self.send_header("content-length", str(len(self.serve_data)))
        self.end_headers()
        self.wfile.write(self.serve_data)


class TestClient(unittest.TestCase):
    script_path: pathlib.Path
    tempdir_path: pathlib.Path
    torrent_path: pathlib.Path
    pty_master: Any
    pty_slave: Any
    torrent: tdummy.Torrent

    def setUp(self) -> None:
        if not sys.platform.startswith("linux"):
            return

        # path relative to this file
        self.script_path = pathlib.Path(__file__).parent / ".." / "client.py"

        self.tempdir = tempfile.TemporaryDirectory()
        self.tempdir_path = pathlib.Path(self.tempdir.name)

        self.torrent = tdummy.get_default()

        # set up a web seed to serve dummy torrent data
        self.server = http.server.HTTPServer(
            ("127.0.0.1", 0), functools.partial(Handler, serve_data=self.torrent.data)
        )
        addr = self.server.server_address[0]
        port = self.server.server_address[1]
        self.server_thread = threading.Thread(target=self.server.serve_forever)
        self.server_thread.start()

        # construct a .torrent to point to our web seed
        tdict = dict(self.torrent.dict)
        tdict[b"url-list"] = b"".join(
            [b"http://", str(addr).encode(), b":", str(port).encode()]
        )
        self.tdict = cast("_Entry", tdict)
        self.torrent_path = self.tempdir_path / "dummy.torrent"
        self.torrent_path.write_bytes(lt.bencode(self.tdict))

        self.pty_master, self.pty_slave = pty.openpty()

    def tearDown(self) -> None:
        if sys.platform != "linux":
            return
        self.server.shutdown()
        self.server.server_close()
        self.tempdir.cleanup()
        os.close(self.pty_master)
        os.close(self.pty_slave)

    @unittest.skipIf(sys.platform != "linux", "pty control only works on linux")
    def test_download_from_web_seed(self) -> None:
        proc = subprocess.Popen(
            [
                sys.executable,
                str(self.script_path),
                "--port",
                "0",
                "--listen-interface",
                "127.0.0.1",
                str(self.torrent_path),
            ],
            cwd=self.tempdir_path,
            stdin=self.pty_slave,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )

        try:
            # wait for the file to be downloaded from the web seed
            file_path = self.tempdir_path / os.fsdecode(self.torrent.files[0].path)
            for _ in lib.loop_until_timeout(15):
                try:
                    if file_path.read_bytes() == self.torrent.data:
                        break
                except FileNotFoundError:
                    pass

            # send a 'q' command
            os.write(self.pty_master, b"q")
            # process should terminate on its own
            returncode = proc.wait(15)
        finally:
            proc.kill()

        # process should complete without errors or warnings
        self.assertEqual(returncode, 0)
        assert proc.stderr is not None  # helps mypy
        self.assertEqual(proc.stderr.read(), "")
        proc.stderr.close()

        # fastresume should be written
        fastresume = pathlib.Path(f"{file_path}.fastresume")
        lt.bdecode(fastresume.read_bytes())
